@@ -1,26 +1,24 @@
 import logging
+import pytz
 from datetime import datetime
 import hashlib
 import os
 import traceback
-from django.core.mail import send_mail
 import math
+
+from django.core.mail import send_mail
 from requests import ConnectionError
-from restkit import RequestFailed
 import simplejson
 from gevent import socket
 import rawes
 import gevent
 from django.conf import settings
+
 from dimagi.utils.decorators.memoized import memoized
-import couchdbkit
+from pillowtop.couchdb import CachedCouchDB
+from .utils import import_settings
 
-if couchdbkit.version_info < (0, 6, 0):
-    USE_NEW_CHANGES = False
-else:
-    from couchdbkit.changes import ChangesStream
-
-    USE_NEW_CHANGES = True
+from couchdbkit.changes import ChangesStream
 
 
 pillow_logging = logging.getLogger("pillowtop")
@@ -28,8 +26,9 @@ pillow_logging.setLevel(logging.INFO)
 
 CHECKPOINT_FREQUENCY = 100
 WAIT_HEARTBEAT = 10000
-RETRY_INTERVAL = 2 #seconds, exponentially increasing
-MAX_RETRIES = 4 #exponential factor threshold for alerts
+CHANGES_TIMEOUT = 60000
+RETRY_INTERVAL = 2  # seconds, exponentially increasing
+MAX_RETRIES = 4  # exponential factor threshold for alerts
 
 INDEX_REINDEX_SETTINGS = {"index": {"refresh_interval": "900s",
                                     "merge.policy.merge_factor": 20,
@@ -46,7 +45,7 @@ INDEX_STANDARD_SETTINGS = {"index": {"refresh_interval": "1s",
 
 
 import functools
-import logging
+
 
 class autoretry_connection(object):
     """
@@ -78,7 +77,6 @@ class autoretry_connection(object):
         return decorated
 
 
-
 class PillowtopIndexingError(Exception):
     pass
 
@@ -99,26 +97,19 @@ class BasicPillow(object):
     extra_args = {} # filter args if needed
     document_class = None  # couchdbkit Document class
     changes_seen = 0
+    couch_db = None
 
-    @property
-    def couch_db(self):
-        return self.document_class.get_db()
+    def __init__(self, couch_db=None, document_class=None):
+        if document_class:
+            self.document_class = document_class
 
-    def old_changes(self):
-        """
-        Couchdbkit < 0.6.0 changes feed listener
-        http://couchdbkit.org/docs/changes_consumer.html
-        """
-        from couchdbkit import Consumer
+        if couch_db:
+            self.couch_db = couch_db
 
-        c = Consumer(self.couch_db, backend='gevent')
-        while True:
-            try:
-                c.wait(self.parsing_processor, since=self.since, filter=self.couch_filter,
-                       heartbeat=WAIT_HEARTBEAT, feed='continuous', timeout=30000, **self.extra_args)
-            except Exception, ex:
-                pillow_logging.exception("Exception in form listener: %s, sleeping and restarting" % ex)
-                gevent.sleep(RETRY_INTERVAL)
+        if not self.couch_db:
+            self.couch_db = self.document_class.get_db() if self.document_class else None
+
+        self.settings = import_settings()
 
     def new_changes(self):
         """
@@ -126,31 +117,30 @@ class BasicPillow(object):
         http://couchdbkit.org/docs/changes.html
         """
         with ChangesStream(self.couch_db, feed='continuous', heartbeat=True, since=self.since,
-                           filter=self.couch_filter, **self.extra_args) as st:
+                           filter=self.couch_filter, include_docs=True, **self.extra_args) as st:
             for c in st:
                 self.processor(c)
 
-    def run(self, since=0):
+    def run(self):
         """
         Couch changes stream creation
         """
         pillow_logging.info("Starting pillow %s" % self.__class__)
+        self.new_changes()
 
-        if USE_NEW_CHANGES:
-            self.new_changes()
-        else:
-            self.old_changes()
-
-    def _get_os_name(self):
-        os_name = "unknown_os"
-        if hasattr(os, "uname"):
+    def _get_machine_id(self):
+        if hasattr(self.settings, 'PILLOWTOP_MACHINE_ID'):
+            os_name = getattr(self.settings, 'PILLOWTOP_MACHINE_ID')
+        elif hasattr(os, 'uname'):
             os_name = os.uname()[1].replace('.', '_')
+        else:
+            os_name = 'unknown_os'
         return os_name
 
     @memoized
     def get_name(self):
         return "%s.%s.%s" % (
-            self.__module__, self.__class__.__name__, self._get_os_name())
+            self.__module__, self.__class__.__name__, self._get_machine_id())
 
     def get_checkpoint_doc_name(self):
         return "pillowtop_%s" % self.get_name()
@@ -161,8 +151,8 @@ class BasicPillow(object):
         if self.couch_db.doc_exist(doc_name):
             checkpoint_doc = self.couch_db.open_doc(doc_name)
         else:
-            #legacy check
-            #split doc and see if non_hostname setup exists.
+            # legacy check
+            # split doc and see if non_hostname setup exists.
             legacy_name = '.'.join(doc_name.split('.')[0:-1])
             starting_seq = "0"
             if self.couch_db.doc_exist(legacy_name):
@@ -184,6 +174,7 @@ class BasicPillow(object):
         checkpoint_doc = self.get_checkpoint()
         checkpoint_doc['old_seq'] = checkpoint_doc['seq']
         checkpoint_doc['seq'] = "0"
+        checkpoint_doc['timestamp'] = datetime.now(tz=pytz.UTC).isoformat()
         self.couch_db.save_doc(checkpoint_doc)
 
     @property
@@ -193,9 +184,16 @@ class BasicPillow(object):
 
     def set_checkpoint(self, change):
         checkpoint = self.get_checkpoint()
+        pillow_logging.info(
+            "(%s) setting checkpoint: %s" % (checkpoint['_id'], change['seq'])
+        )
         checkpoint['seq'] = change['seq']
+        checkpoint['timestamp'] = datetime.now(tz=pytz.UTC).isoformat()
         self.couch_db.save_doc(checkpoint)
 
+    def get_db_seq(self):
+        return self.couch_db.info()['update_seq']
+
     def parsing_processor(self, change):
         """
         Processor that also parses the change to json - only for pre 0.6.0 couchdbkit,
@@ -203,17 +201,7 @@ class BasicPillow(object):
         """
         self.processor(simplejson.loads(change))
 
-    def processor(self, change, do_set_checkpoint=True):
-        """
-        Parent processsor for a pillow class - this should not be overridden.
-        This workflow is made for the situation where 1 change yields 1 transport/transaction
-        """
-        self.changes_seen += 1
-        if self.changes_seen % CHECKPOINT_FREQUENCY == 0 and do_set_checkpoint:
-            pillow_logging.info(
-                "(%s) setting checkpoint: %s" % (self.get_checkpoint_doc_name(), change['seq']))
-            self.set_checkpoint(change)
-
+    def process_change(self, change):
         try:
             t = self.change_trigger(change)
             if t is not None:
@@ -221,9 +209,19 @@ class BasicPillow(object):
                 if tr is not None:
                     self.change_transport(tr)
         except Exception, ex:
-            pillow_logging.exception("Error on change: %s, %s" % (change['id'], ex))
-            gevent.sleep(RETRY_INTERVAL)
+            pillow_logging.exception("[%s] Error on change: %s, %s" % (self.get_name(), change['id'], ex))
 
+    def processor(self, change, do_set_checkpoint=True):
+        """
+        Parent processsor for a pillow class - this should not be overridden.
+        This workflow is made for the situation where 1 change yields 1 transport/transaction
+        """
+
+        self.changes_seen += 1
+        if self.changes_seen % CHECKPOINT_FREQUENCY == 0 and do_set_checkpoint:
+            self.set_checkpoint(change)
+
+        self.process_change(change)
 
     def change_trigger(self, changes_dict):
         """
@@ -239,7 +237,10 @@ class BasicPillow(object):
         if changes_dict.get('deleted', False):
             #override deleted behavior on consumers that care/deal with deletions
             return None
-        return self.couch_db.open_doc(changes_dict['id'])
+        if 'doc' in changes_dict:
+            return changes_dict['doc']
+        else:
+            return self.couch_db.open_doc(changes_dict['id'])
 
     def change_transform(self, doc_dict):
         """
@@ -257,6 +258,61 @@ class BasicPillow(object):
             "Error, this pillowtop subclass has not been configured to do anything!")
 
 
+class PythonPillow(BasicPillow):
+    """
+    A pillow that does filtering in python instead of couch.
+
+    Useful because it will actually set checkpoints throughout even if there
+    are no matched docs.
+
+    In initial profiling this was also 2-3x faster than the couch-filtered
+    version.
+
+    Subclasses should override the python_filter function to perform python
+    filtering.
+    """
+    CHUNK_SIZE = 250
+    PILLOW_CHECKPOINT_FREQUENCY = CHECKPOINT_FREQUENCY * 10
+
+    def __init__(self, document_class=None):
+        super(PythonPillow, self).__init__(document_class=document_class)
+
+        if self.document_class:
+            self.couch_db = CachedCouchDB(self.document_class.get_db().uri, readonly=False)
+
+    def python_filter(self, doc):
+        """
+        Should return True if the doc is to be processed by your pillow
+        """
+        return True
+
+    def process_chunk(self):
+        self.couch_db.bulk_load([change['id'] for change in self.change_queue],
+                                    purge_existing=True)
+
+        while len(self.change_queue) > 0:
+            change = self.change_queue.pop()
+            doc = self.couch_db.open_doc(change['id'], check_main=False)
+            if doc and self.python_filter(doc):
+                try:
+                    self.process_change(change)
+                except Exception:
+                    logging.exception('something went wrong processing change %s (%s)' % (change['seq'], change['id']))
+
+
+    def processor(self, change, do_set_checkpoint=True):
+        self.changes_seen += 1
+        self.change_queue.append(change)
+        if self.changes_seen % self.PILLOW_CHECKPOINT_FREQUENCY == 0 and do_set_checkpoint:
+            self.set_checkpoint(change)
+        if len(self.change_queue) > self.CHUNK_SIZE:
+            self.process_chunk()
+
+    def run(self):
+        self.change_queue = []
+        super(PythonPillow, self).run()
+
+
 class ElasticPillow(BasicPillow):
     """
     Elasticsearch handler
@@ -274,10 +330,11 @@ class ElasticPillow(BasicPillow):
     # index to always have the latest version of the case based upon ALL changes done to it.
     allow_updates = True
 
-    def __init__(self, create_index=True, online=True):
+    def __init__(self, create_index=True, online=True, **kwargs):
         """
         create_index if the index doesn't exist on the ES cluster
         """
+        super(ElasticPillow, self).__init__(**kwargs)
         self.online = online
         index_exists = self.index_exists()
         if create_index and not index_exists:
@@ -350,6 +407,8 @@ class ElasticPillow(BasicPillow):
                 pillow_logging.error("ElasticPillow: error deleting route %s - ignoring: %s" % \
                               (self.get_doc_path(changes_dict['id']), ex))
             return None
+        if 'doc' in changes_dict:
+            return changes_dict['doc']
         else:
             return self.couch_db.open_doc(changes_dict['id'])
 
@@ -387,7 +446,7 @@ class ElasticPillow(BasicPillow):
                                       "_id": tr['_id']}}
                         yield tr
             except Exception, ex:
-                pillow_logging.error("Error on change: %s, %s" % (change['id'], ex))
+                pillow_logging.error("[%s] Error on change: %s, %s" % (self.get_name(), change['id'], ex))
 
     def process_bulk(self, changes):
         self.allow_updates = False
@@ -396,11 +455,11 @@ class ElasticPillow(BasicPillow):
         bstart = datetime.utcnow()
         bulk_payload = '\n'.join(map(simplejson.dumps, self.bulk_builder(changes))) + "\n"
         pillow_logging.info(
-            "prepare_bulk,%s" % str(ms_from_timedelta(datetime.utcnow() - bstart) / 1000.0))
+            "%s,prepare_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - bstart) / 1000.0)))
         send_start = datetime.utcnow()
         es.post('_bulk', data=bulk_payload)
         pillow_logging.info(
-            "send_bulk,%s" % str(ms_from_timedelta(datetime.utcnow() - send_start) / 1000.0))
+            "%s,send_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - send_start) / 1000.0)))
 
     def processor(self, change, do_set_checkpoint=True):
         """
@@ -409,18 +468,9 @@ class ElasticPillow(BasicPillow):
         """
         self.changes_seen += 1
         if self.changes_seen % CHECKPOINT_FREQUENCY == 0 and do_set_checkpoint:
-            pillow_logging.info(
-                "(%s) setting checkpoint: %s" % (self.get_checkpoint_doc_name(), change['seq']))
             self.set_checkpoint(change)
 
-        try:
-            t = self.change_trigger(change)
-            if t is not None:
-                tr = self.change_transform(t)
-                if tr is not None:
-                    self.change_transport(tr)
-        except Exception, ex:
-            pillow_logging.error("Error on change: %s, %s" % (change['id'], ex))
+        self.process_change(change)
 
     def send_robust(self, path, data={}, retries=MAX_RETRIES, except_on_failure=False, update=False):
         """
@@ -628,7 +678,7 @@ class AliasedElasticPillow(ElasticPillow):
         class name and the hashed name representation.
         """
         return "%s.%s.%s.%s" % (
-            self.__module__, self.__class__.__name__, self.calc_meta(), self._get_os_name())
+            self.__module__, self.__class__.__name__, self.calc_meta(), self._get_machine_id())
 
     def get_mapping_from_type(self, doc_dict):
         raise NotImplementedError("This must be implemented in this subclass!")
@@ -721,4 +771,3 @@ class LogstashMonitoringPillow(NetworkPillow):
             return {}
         else:
             return super(NetworkPillow, self).processor(change)
-
